#include <stdio.h>
#include <stdlib.h>
#include "viglib.c"

float ic(char *s, size_t len, short key_len)
{
 unsigned int freq[256], i, j;
 float ic = 0.0;

 for(j=0; j<256; j++) freq[j]=0;
 for(j=0; key_len*j<len; j++) freq[s[key_len*j]]++;
 for(i=0; i<256; i++) ic += freq[i]*(freq[i]-1);

 return (ic/(j*(j-1)));
}

float icm(char *s1, size_t len1, char *s2, size_t len2)
{
 unsigned int freq1[256], freq2[256], i;
 float icm = 0.0;
 for(i=0; i<256; i++) freq1[i]=freq2[i]=0;
 for(i=0; i<len1; i++) freq1[s1[i]]++;
 for(i=0; i<len2; i++) freq2[s2[i]]++;
 for(i=0; i<256; i++) icm += freq1[i]*freq2[i];
 return (icm/(len1*len2));
}

void* find_key(char* key, short klen, char* s, int len)
{
 unsigned int freq1[256], freq2[256], i,j,k,l;
 float icm;
 char c;

 for(l=0; l<256; l++) freq1[l]=0;
 for(j=0; klen*j<len; j++) freq1[s[klen*j]]++;
 key[0] = 'a';

 for(i=1; i<klen; i++, icm=0.0)
 {
  for(c=-26; c<26; c++)
  {
   float ic = 0.0;
   for(l=0; l<256; l++) freq2[l]=0;
   for(k=0; i+klen*k<len; k++)
   {
     unsigned char p = s[i+klen*k]-c;
     freq2[p]++;
   }
   for(l=0; l<256; l++) ic += freq1[l]*freq2[l];
   ic /= (j*k);
   if(ic >= icm){icm=ic;key[i]=(c>=0 ? 'a'+c : 'z'+c+1);}
  }
 }
}

bool test_key_size(char *key, int klen, int ks)
{
 int i,j;
 if(klen%ks != 0) return false;
 for(i=1; i < (klen/ks); i++)
 {
  for(j=0; j<ks; j++) if(key[j+ks*i] != key[j]) return false;
 }
 return true;
}

int main(int argc, char **argv)
{
 int i, j, key_len, max_size;
 float c, d;
 size_t len;
 FILE *fd, *fref;

 if(argc < 2)
  return printf("Usage : %s <in_file> [max_key_length] [<ref_file>]\n", argv[0]);

 if(argc < 3)
 {
  printf("Warning : max_size omitted, using default (50)\n");
  max_size = 50;
 }
 else
 {
  sscanf(argv[2], "%d", &max_size);
  if(max_size < 1 || max_size > 50)
  {
   printf("Warning : invalid max_size, defaulting to 50\n");
   max_size = 50;
  }
 }

 if(argc >= 4)
 {
  fref = fopen(argv[3], "rb");
  if(fref == NULL) return printf("Error : can't open <%s> for reading !\n", argv[3]);
 }

 fd = fopen(argv[1], "rb");
 if(fd == NULL) return printf("Error : can't open <%s> for reading !\n", argv[1]);

 char *buf = malloc(1024*1024);
 len = fread(buf, 1, 1024*1024, fd);

 for(c=0.0, i=1; i<=max_size && i<len; i++)
  if((d=ic(buf, len, i))>=c){ c=d; key_len=i; }

 char *key = malloc(key_len+1);
 find_key(key, key_len, buf, len);
 for(i=1; i<key_len; i++) if(test_key_size(key,key_len,i)){ key_len=i; break;}
 key[key_len]=0;

 if(20 * key_len > len)
  printf("Warning : expected key size is relatively large.\nTry to decrease [max_size] if results are wrong.\n");

 if(argc < 4)
 {
  printf("+--------"); for(i=0;i<key_len;i++) printf("---"); printf("+\n| ");
  for(i=0; i<26; i++){
   printf("%s", key); if((i+1)%3) printf(" | "); else printf(" |\n| ");
   for(j=0; j < key_len; j++)
   {
    key[j]++; if(key[j]>'z') key[j]='a';
   }
  }
  for(i=0;i<key_len;i++) printf(" "); printf(" |\n+");
  for(i=0;i<key_len;i++) printf("---"); printf("--------+\n");
 }
 else
 {
  char *ref = malloc(1024*1024);
  char *dec = malloc(len);
  char *good_key = malloc(key_len+1);
  size_t rlen = fread(ref, 1, 1024*1024, fref);

  for(c=0.0, i=0; i<26; i++)
  {
   memcpy(dec,buf,len);
   vigenere(dec,len,key,true);
   d = icm(ref,rlen,dec,len);
   if(d >= c){c=d; strcpy(good_key,key);}

   for(j=0; j < key_len; j++)
   {
    key[j]++; if(key[j]>'z') key[j]='a';
   }
  }
  printf("Most likely key is %s\n",good_key);
  free(ref); free(dec); free(good_key);
  fclose(fref);
 }


 fclose(fd); free(buf); free(key);
 return 0;
}
